"""Module providing translation from wxPython events to OpenGLContext events"""
from OpenGLContext.events import mouseevents, keyboardevents, eventhandlermixin
from wxPython.wx import *
import time

class EventHandlerMixin( eventhandlermixin.EventHandlerMixin):
	"""wxPython-specific event handler mix-in

	Basically provides translation from wxPython events
	to their OpenGLContext-specific equivalents (the
	concrete versions of which are also defined in this
	module).
	"""
	### KEYBOARD interactions
	def wxOnKeyDown( self, event ):
		'''Convert a key-press to a context-style event'''
		self.ProcessEvent( wxKeyboardEvent( self, event, 1))
		event.Skip()
	def wxOnKeyUp( self, event ):
		'''Convert a key-release to a context-style event'''
		self.ProcessEvent( wxKeyboardEvent( self, event, 0))
	def wxOnCharacter( self, event ):
		"""Convert character (non-control) press to context event"""
		self.ProcessEvent( wxKeypressEvent( self, event))
	### MOUSE Interaction
	def wxOnMouseButton(self, event ):
		"""Convert mouse-button event to context event"""
		self.addPickEvent( wxMouseButtonEvent( self, event))
		self.triggerPick()
	def wxOnMouseMove(self, event ):
		"""Convert mouse-movement event to context event"""
		self.addPickEvent( wxMouseMoveEvent( self, event))
		self.triggerPick()

class wxXEvent(object):
	"""Base-class for all wxPython-specific event classes

	Provides method for determining the modifier set from
	wxPython event objects
	"""
	def _getModifiers( self, wxEventObject):
		"""Get a three-tupple of shift, control, alt status"""
		return (
			not(not( wxEventObject.ShiftDown())),
			not(not( wxEventObject.ControlDown())),
			not(not( wxEventObject.AltDown())),
		)

class wxMouseButtonEvent( wxXEvent, mouseevents.MouseButtonEvent ):
	"""wxPython-specific mouse button event"""
	def __init__( self, context, wxEventObject ):
		super (wxMouseButtonEvent, self).__init__()
		if hasattr( context, 'currentPass'):
			self.renderingPass = context.currentPass
		self.modifiers = self._getModifiers(wxEventObject)
##		import pdb
##		pdb.set_trace()
		for local, wx in ( (0,1), (1,3), (2,2)):
			if wxEventObject.Button( wx ):
				self.button = local
				self.state = wxEventObject.ButtonDown()
				break
		self.pickPoint = wxEventObject.GetX(), context.getViewPort()[1]- wxEventObject.GetY()
		
class wxMouseMoveEvent( wxXEvent, mouseevents.MouseMoveEvent ):
	"""wxPython-specific mouse movement event"""
	def __init__( self, context, wxEventObject ):
		super (wxMouseMoveEvent, self).__init__()
		if hasattr( context, 'currentPass'):
			self.renderingPass = context.currentPass
		self.modifiers = self._getModifiers(wxEventObject)
		buttons = []
		for local, method in ( (0,"LeftIsDown"), (1,"MiddleIsDown"), (2,"RightIsDown")):
			if getattr( wxEventObject, method )():
				buttons.append( local )
		self.buttons = tuple( buttons )
		self.pickPoint = wxEventObject.GetX(), context.getViewPort()[1]- wxEventObject.GetY()

class wxKeyboardEvent( wxXEvent, keyboardevents.KeyboardEvent ):
	"""wxPython-specific keyboard event"""
	def __init__( self, context, wxEventObject, state=0 ):
		super (wxKeyboardEvent, self).__init__()
		if hasattr( context, 'currentPass'):
			self.renderingPass = context.currentPass
		self.modifiers = self._getModifiers(wxEventObject)
		self.name = keyboardMapping.get( wxEventObject.GetKeyCode())	
		self.state = state
		
class wxKeypressEvent( wxXEvent, keyboardevents.KeypressEvent ):
	"""wxPython-specific key-press event"""
	def __init__( self, context, wxEventObject):
		super (wxKeypressEvent, self).__init__()
		if hasattr( context, 'currentPass'):
			self.renderingPass = context.currentPass
		self.modifiers = self._getModifiers(wxEventObject)
		self.name = keyboardMapping.get( wxEventObject.GetKeyCode())	

keyboardMapping = {
	WXK_BACK:'<back>',
	WXK_TAB:'<tab>',
	WXK_RETURN:'<return>',
	WXK_ESCAPE:'<escape>',
	WXK_SPACE:' ',
	WXK_DELETE:'<delete>',
	WXK_START:'<start>',
##	WXK_LBUTTON:
##	WXK_RBUTTON:
##	WXK_CANCEL:
##	WXK_MBUTTON:
##	WXK_CLEAR:
	WXK_SHIFT:'<shift>',
	WXK_CONTROL:'<ctrl>',
	WXK_MENU:'<alt>',
	WXK_PAUSE:'<pause>',
##	WXK_CAPITAL:
	WXK_PRIOR:'<pageup>',
	WXK_NEXT:'<pagedown>',
	WXK_END:'<end>',
	WXK_HOME:'<home>',
	WXK_LEFT:'<left>',
	WXK_UP:'<up>',
	WXK_RIGHT:'<right>',
	WXK_DOWN:'<down>',
##	WXK_SELECT:
##	WXK_PRINT:
##	WXK_EXECUTE:
##	WXK_SNAPSHOT:
	WXK_INSERT:'<insert>',
##	WXK_HELP:
	WXK_NUMPAD0:'#0',
	WXK_NUMPAD1:'#1',
	WXK_NUMPAD2:'#2',
	WXK_NUMPAD3:'#3',
	WXK_NUMPAD4:'#4',
	WXK_NUMPAD5:'#5',
	WXK_NUMPAD6:'#6',
	WXK_NUMPAD7:'#7',
	WXK_NUMPAD8:'#8',
	WXK_NUMPAD9:'#9',
	WXK_MULTIPLY:'*',
	WXK_ADD:'+',
	WXK_SEPARATOR:'_', # ???
	WXK_SUBTRACT:'-',
	WXK_DECIMAL:'.',
	WXK_DIVIDE:'/',
	WXK_F1: '<F1>',
	WXK_F2: '<F2>',
	WXK_F3: '<F3>',
	WXK_F4: '<F4>',
	WXK_F5: '<F5>',
	WXK_F6: '<F6>',
	WXK_F7: '<F7>',
	WXK_F8: '<F8>',
	WXK_F9: '<F9>',
	WXK_F10: '<F10>',
	WXK_F11: '<F11>',
	WXK_F12: '<F12>',
	WXK_F13: '<F13>',
	WXK_F14: '<F14>',
	WXK_F15: '<F15>',
	WXK_F16: '<F16>',
	WXK_F17: '<F17>',
	WXK_F18: '<F18>',
	WXK_F19: '<F19>',
	WXK_F20: '<F20>',
	WXK_F21: '<F21>',
	WXK_F22: '<F22>',
	WXK_F23: '<F23>',
	WXK_F24: '<F24>',
	WXK_NUMLOCK: '<numlock>',
	WXK_SCROLL: '<scroll>',
	13: '<return>',
}
for integer in range (256):
	if not keyboardMapping.has_key( integer ):
		keyboardMapping[integer] = chr( integer ).lower()
	